@@ -1,6 +1,6 @@
-;;; wisi-parse.el --- Wisi parser
+;;; wisi-parse.el --- Wisi parser  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2013, 2014  Free Software Foundation, Inc.
+;; Copyright (C) 2013-2015  Free Software Foundation, Inc.
 
 ;; This file is part of GNU Emacs.
 
@@ -95,6 +95,12 @@ point at which that max was spawned.")
 (defvar-local wisi-cache-max 0
   "Maximimum position in buffer where wisi-cache text properties are valid.")
 
+(defun wisi-token-text (token)
+  "Return buffer text from token range."
+  (let ((region (cdr token)))
+    (and region
+       (buffer-substring-no-properties (car region) (cdr region)))))
+
 (defun wisi-parse (automaton lexer)
   "Parse current buffer from bob using the automaton specified in AUTOMATON.
 
@@ -105,6 +111,7 @@ point at which that max was spawned.")
   list (symbol text start . end), where `symbol' is the terminal
   symbol, `text' is the token string, `start . end' is the range
   in the buffer."
+
   (let* ((actions (aref automaton 0))
 	 (gotos   (aref automaton 1))
 	 (parser-states ;; vector of parallel parser states
@@ -142,8 +149,8 @@ point at which that max was spawned.")
 			(let ((state (aref (wisi-parser-state-stack parser-state)
 					   (wisi-parser-state-sp parser-state))))
 			  (wisi-error-msg (concat "too many parallel parsers required in grammar state %d;"
-						  " simplify grammar, or increase `wisi-parse-max-parallel'"
-						  state)))))
+						  " simplify grammar, or increase `wisi-parse-max-parallel'")
+						  state))))
 
 	      (let ((j (wisi-free-parser parser-states)))
 		(cond
@@ -188,7 +195,7 @@ point at which that max was spawned.")
 		     (signal 'wisi-parse-error
 			     (wisi-error-msg "syntax error in grammar state %d; unexpected %s, expecting one of %s"
 					     state
-					     (nth 1 token)
+					     (wisi-token-text token)
 					     (mapcar 'car (aref actions state))))
 		     ))
 		  (t
@@ -205,7 +212,7 @@ point at which that max was spawned.")
 					 (wisi-error-msg
 					  "syntax error in grammar state %d; unexpected %s, expecting one of %s"
 					  state
-					  (nth 1 token)
+					  (wisi-token-text token)
 					  (mapcar 'car (aref actions state)))))
 			   )))
 		     (signal 'wisi-parse-error msg)))
@@ -213,11 +220,11 @@ point at which that max was spawned.")
 
 		(1
 		 (setf (wisi-parser-state-active parser-state) nil); Don't save error for later.
-		 (wisi-execute-pending (wisi-parser-state-pending
-					(aref parser-states (wisi-active-parser parser-states))))
-		 (setf (wisi-parser-state-pending
-			(aref parser-states (wisi-active-parser parser-states)))
-		       nil))
+		 (let ((parser-state (aref parser-states (wisi-active-parser parser-states))))
+		   (wisi-execute-pending (wisi-parser-state-label parser-state)
+					 (wisi-parser-state-pending parser-state))
+		   (setf (wisi-parser-state-pending parser-state) nil)
+		   ))
 		(t
 		 ;; We were in a parallel parse, and this parser
 		 ;; failed; mark it inactive, don't save error for
@@ -260,20 +267,20 @@ point at which that max was spawned.")
 (defun wisi-parsers-active (parser-states active-count)
   "Return the type of parser cycle to execute.
 PARSER-STATES[*].active is the last action a parser took. If it
-was 'shift, that parser used the input token, and should not be
+was `shift', that parser used the input token, and should not be
 executed again until another input token is available, after all
 parsers have shifted the current token or terminated.
 
 Returns one of:
 
-'accept : all PARSER-STATES have active set to nil or 'accept -
+`accept' : all PARSER-STATES have active set to nil or `accept' -
 done parsing
 
-'shift : all PARSER-STATES have active set to nil, 'accept, or
-'shift - get a new token, execute 'shift parsers.
+`shift' : all PARSER-STATES have active set to nil, `accept', or
+`shift' - get a new token, execute `shift' parsers.
 
-'reduce : some PARSER-STATES have active set to 'reduce - no new
-token, execute 'reduce parsers."
+`reduce' : some PARSER-STATES have active set to `reduce' - no new
+token, execute `reduce' parsers."
   (let ((result nil)
 	(i 0)
 	(shift-count 0)
@@ -327,7 +334,7 @@ token, execute 'reduce parsers."
 (defun wisi-parse-elim-identical (parser-states active-parser-count)
   "Check for parsers in PARSER-STATES that have reached identical states eliminate one.
 Return new ACTIVE-PARSER-COUNT. Assumes all parsers have active
-nil, 'shift, or 'accept."
+nil, `shift', or `accept'."
   ;; parser-states passed by reference; active-parser-count by copy
   ;; see test/ada_mode-slices.adb for example
   (dotimes (parser-i (1- (length parser-states)))
@@ -340,7 +347,7 @@ nil, 'shift, or 'accept."
 	      (dotimes (stack-i (wisi-parser-state-sp (aref parser-states parser-i)))
 		(setq
 		 compare
-		 (and compare
+		 (and compare ;; bypass expensive 'arefs' after first stack item compare fail
 		      (equal (aref (wisi-parser-state-stack (aref parser-states parser-i)) stack-i)
 			     (aref (wisi-parser-state-stack (aref parser-states (+ parser-i parser-j 1))) stack-i)))))
 	      (when compare
@@ -349,19 +356,18 @@ nil, 'shift, or 'accept."
 		(when (> wisi-debug 1)
 		  (message "terminate identical parser %d (%d active)"
 			   (+ parser-i parser-j 1) active-parser-count))
+		(setf (wisi-parser-state-active (aref parser-states (+ parser-i parser-j 1))) nil)
 		(when (= active-parser-count 1)
 		  ;; the actions for the two parsers are not
 		  ;; identical, but either is good enough for
-		  ;; indentation and navigation, so we just do one.
-		  (when (> wisi-debug 1) (message "executing actions for %d" (+ parser-i parser-j 1)))
-		  (wisi-execute-pending (wisi-parser-state-pending (aref parser-states (+ parser-i parser-j 1))))
-		  (setf (wisi-parser-state-pending (aref parser-states (+ parser-i parser-j 1))) nil)
-
-		  ;; clear pending of other parser so it can be reused
-		  (setf (wisi-parser-state-pending (aref parser-states parser-i)) nil))
-
-		(setf (wisi-parser-state-active (aref parser-states (+ parser-i parser-j 1))) nil))
-	      )))
+		  ;; indentation and navigation, so we just do the
+		  ;; actions for the one that is not terminating.
+		  (let ((parser-state (aref parser-states parser-i)))
+		    (wisi-execute-pending (wisi-parser-state-label parser-state)
+					  (wisi-parser-state-pending parser-state))
+		    (setf (wisi-parser-state-pending parser-state) nil)
+		    ))
+		))))
 	)))
   active-parser-count)
 
@@ -370,31 +376,39 @@ nil, 'shift, or 'accept."
   (let ((result (if tokens 0 (point))))
     (mapc
      (lambda (token)
-       (when (cl-cdddr token)
-	 (setq result (max (cl-cdddr token) result))))
+       (when (cddr token)
+	 (setq result (max (cddr token) result))))
      tokens)
     result)
   )
 
-(defun wisi-parse-exec-action (func tokens)
+(defun wisi-parse-exec-action (func nonterm tokens)
   "Execute action if all tokens past wisi-cache-max."
   ;; We don't execute actions if all tokens are before wisi-cache-max,
   ;; because later actions can update existing caches, and if the
   ;; parse fails that won't happen. It also saves time.
-  (if (>= (wisi-parse-max-pos tokens) wisi-cache-max)
+  ;;
+  ;; Also skip if no tokens; nothing to do. This can happen when all
+  ;; tokens in a grammar statement are optional.
+  (if (< 0 (length tokens))
+      (if (>= (wisi-parse-max-pos tokens) wisi-cache-max)
 
-      (funcall func tokens)
+	  (funcall func nonterm tokens)
+
+	(when (> wisi-debug 1)
+	  (message "... action skipped; before wisi-cache-max %d" wisi-cache-max)))
 
     (when (> wisi-debug 1)
-      (message "... action skipped"))
+      (message "... action skipped; no tokens"))
     ))
 
-(defun wisi-execute-pending (pending)
+(defun wisi-execute-pending (parser-label pending)
+  (when (> wisi-debug 1) (message "%d: pending actions:" parser-label))
   (while pending
     (when (> wisi-debug 1) (message "%s" (car pending)))
 
     (let ((func-args (pop pending)))
-      (wisi-parse-exec-action (car func-args) (cadr func-args)))
+      (wisi-parse-exec-action (nth 0 func-args) (nth 1 func-args) (cl-caddr func-args)))
     ))
 
 (defun wisi-parse-1 (token parser-state pendingp actions gotos)
@@ -475,18 +489,18 @@ Return nil."
   "Return a pair (START . END), the buffer region for a nonterminal.
 STACK is the parser stack.  I and J are the indices in STACK of
 the first and last tokens of the nonterminal."
-  (let ((start (cl-caddr (aref stack i)))
-        (end   (cl-cdddr (aref stack j))))
+  (let ((start (cadr (aref stack i)))
+        (end   (cddr (aref stack j))))
     (while (and (or (not start) (not end))
 		(/= i j))
       (cond
        ((not start)
 	;; item i is an empty production
-	(setq start (cl-caddr (aref stack (setq i (+ i 2))))))
+	(setq start (cadr (aref stack (setq i (+ i 2))))))
 
        ((not end)
 	;; item j is an empty production
-	(setq end (cl-cdddr (aref stack (setq j (- j 2))))))
+	(setq end (cddr (aref stack (setq j (- j 2))))))
 
        (t (setq i j))))
     (and start end (cons start end))))
@@ -495,34 +509,40 @@ the first and last tokens of the nonterminal."
   "Reduce PARSER-STATE.stack, and execute or pend ACTION."
   (let* ((stack (wisi-parser-state-stack parser-state)); reference
 	 (sp (wisi-parser-state-sp parser-state)); copy
-	 (token-count (or (nth 2 action) 0))
+	 (token-count (nth 2 action))
 	 (nonterm (nth 0 action))
 	 (nonterm-region (when (> token-count 0)
 			   (wisi-nonterm-bounds stack (- sp (* 2 (1- token-count)) 1) (1- sp))))
 	 (post-reduce-state (aref stack (- sp (* 2 token-count))))
 	 (new-state (cdr (assoc nonterm (aref gotos post-reduce-state))))
-	 tokens)
+	 (tokens (make-vector token-count nil)))
+
     (when (not new-state)
       (error "no goto for %s %d" nonterm post-reduce-state))
-    (if (= 1 token-count)
-	(setq tokens (list (aref stack (1- sp))))
+
+    (when (nth 1 action)
+      ;; don't need wisi-tokens for a null user action
       (dotimes (i token-count)
-	(push (aref stack (- sp (* 2 i) 1)) tokens)))
+	(aset tokens (- token-count i 1) (aref stack (- sp (* 2 i) 1)))))
+
     (setq sp (+ 2 (- sp (* 2 token-count))))
-    (aset stack (1- sp) (cons nonterm (cons nil nonterm-region)))
+    (aset stack (1- sp) (cons nonterm nonterm-region))
     (aset stack sp new-state)
     (setf (wisi-parser-state-sp parser-state) sp)
-    (if pendingp
-	(if (wisi-parser-state-pending parser-state)
-	    (setf (wisi-parser-state-pending parser-state)
-		  (append (wisi-parser-state-pending parser-state)
-			  (list (list (nth 1 action) tokens))))
-	  (setf (wisi-parser-state-pending parser-state)
-		(list (list (nth 1 action) tokens))))
 
-      ;; Not pending.
-      (wisi-parse-exec-action (nth 1 action) tokens)
-      )
+    (when (nth 1 action)
+      ;; nothing to do for a null user action
+      (if pendingp
+	  (if (wisi-parser-state-pending parser-state)
+	      (setf (wisi-parser-state-pending parser-state)
+		    (append (wisi-parser-state-pending parser-state)
+			    (list (list (nth 1 action) nonterm tokens))))
+	    (setf (wisi-parser-state-pending parser-state)
+		  (list (list (nth 1 action) nonterm tokens))))
+
+	;; Not pending.
+	(wisi-parse-exec-action (nth 1 action) nonterm tokens)
+	))
     ))
 
 (provide 'wisi-parse)
